from pathlib import Path

REPO_DIR = Path(__file__).parents[1].absolute()
WHO_TABLES = [
    'wfl_boys_0_2_zscores.json', 'wfl_girls_0_2_zscores.json',
    'wfh_boys_2_5_zscores.json', 'wfh_girls_2_5_zscores.json',
    'wfa_boys_0_5_zscores.json', 'wfa_girls_0_5_zscores.json',
    'lhfa_boys_0_5_zscores.json', 'lhfa_girls_0_5_zscores.json',
]
